
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <CL/cl.h>
#include "clFFT.h"


/*
 * Utility functions
 */
const char* OPENCL_ERROR_MSG[] = {
    "CL_SUCCESS",
    "CL_DEVICE_NOT_FOUND",
    "CL_DEVICE_NOT_AVAILABLE",
    "CL_COMPILER_NOT_AVAILABLE",
    "CL_MEM_OBJECT_ALLOCATION_FAILURE",
    "CL_OUT_OF_RESOURCES",
    "CL_OUT_OF_HOST_MEMORY",
    "CL_PROFILING_INFO_NOT_AVAILABLE",
    "CL_MEM_COPY_OVERLAP",
    "CL_IMAGE_FORMAT_MISMATCH",
    "CL_IMAGE_FORMAT_NOT_SUPPORTED",
    "CL_BUILD_PROGRAM_FAILURE",
    "CL_MAP_FAILURE",
    "CL_MISALIGNED_SUB_BUFFER_OFFSET",
    "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST",

    /* next IDs start at 30! */
    "CL_INVALID_VALUE",
    "CL_INVALID_DEVICE_TYPE",
    "CL_INVALID_PLATFORM",
    "CL_INVALID_DEVICE",
    "CL_INVALID_CONTEXT",
    "CL_INVALID_QUEUE_PROPERTIES",
    "CL_INVALID_COMMAND_QUEUE",
    "CL_INVALID_HOST_PTR",
    "CL_INVALID_MEM_OBJECT",
    "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR",
    "CL_INVALID_IMAGE_SIZE",
    "CL_INVALID_SAMPLER",
    "CL_INVALID_BINARY",
    "CL_INVALID_BUILD_OPTIONS",
    "CL_INVALID_PROGRAM",
    "CL_INVALID_PROGRAM_EXECUTABLE",
    "CL_INVALID_KERNEL_NAME",
    "CL_INVALID_KERNEL_DEFINITION",
    "CL_INVALID_KERNEL",
    "CL_INVALID_ARG_INDEX",
    "CL_INVALID_ARG_VALUE",
    "CL_INVALID_ARG_SIZE",
    "CL_INVALID_KERNEL_ARGS",
    "CL_INVALID_WORK_DIMENSION",
    "CL_INVALID_WORK_GROUP_SIZE",
    "CL_INVALID_WORK_ITEM_SIZE",
    "CL_INVALID_GLOBAL_OFFSET",
    "CL_INVALID_EVENT_WAIT_LIST",
    "CL_INVALID_EVENT",
    "CL_INVALID_OPERATION",
    "CL_INVALID_GL_OBJECT",
    "CL_INVALID_BUFFER_SIZE",
    "CL_INVALID_MIP_LEVEL",
    "CL_INVALID_GLOBAL_WORK_SIZE"
};

static const char* opencl_map_error(cl_int error)
{
    if (error >= -14)
        return OPENCL_ERROR_MSG[-error];
    if (error <= -30)
        return OPENCL_ERROR_MSG[-error-15];
    return NULL;
}

#define CL_CHECK_ERROR(errorcode) if (errorcode != CL_SUCCESS) { \
    printf("OpenCL Error %i <%s:%i>", errorcode, __FILE__, __LINE__); }

#define CL_CHECK_ERROR_EX(errorcode) if (errorcode != CL_SUCCESS) { \
    printf("OpenCL Error: %s <%s:%i>", opencl_map_error(errorcode), __FILE__, __LINE__); }


void print_array(float* a, size_t n)
{
    printf("[ ");
    for (size_t i = 0; i < n-2; i += 2)
        printf("(%2.3f, %2.3f), ", a[i], a[i+1]);
    printf("(%2.3f, %2.3f) ]\n", a[n-2], a[n-1]);
}

int main(int argc, const char *argv[])
{
    cl_platform_id platform;
    CL_CHECK_ERROR(clGetPlatformIDs(1, &platform, NULL));

    cl_device_id device;
    CL_CHECK_ERROR(clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 
                1, &device, NULL));

    cl_int err;
    cl_context context = clCreateContext(0, 1, &device, NULL, NULL, &err);
    CL_CHECK_ERROR(err);
    cl_command_queue queue = clCreateCommandQueue(context, device, 0, &err);
    CL_CHECK_ERROR(err);

    const int N = 8;
    float h_input[] = { 1.0, 0.0, 2.3, 0.0, 1.1, 0.0, 1.0, 0.0, 3.3, 0.0, 2.0, 0.0, 1.0, 0.0, 3.2, 0.0 };
    float h_output[2*N];
    h_output[0] = 0.0;

    clFFT_Dim3 dim;
    dim.x = N; dim.y = 1; dim.z = 1;
    clFFT_Plan fft_plan = clFFT_CreatePlan(context, dim, clFFT_1D, clFFT_InterleavedComplexFormat, &err);
    CL_CHECK_ERROR_EX(err);
    if (fft_plan == NULL) {
        printf("Error creating FFT plan\n");
        return 255;
    }
    FILE *fp = fopen("plan.txt", "w");
    clFFT_DumpPlan(fft_plan, fp);
    fclose(fp);

    cl_mem d_input = clCreateBuffer(context, CL_MEM_READ_WRITE, 2*N*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);
    if (h_input == NULL) {
        printf("nice\n");
    }
    CL_CHECK_ERROR_EX(clEnqueueWriteBuffer(queue, d_input, CL_TRUE, 0, 2*N*sizeof(float), h_input, 0, NULL, NULL));

    cl_mem d_output = clCreateBuffer(context, CL_MEM_READ_WRITE, 2*N*sizeof(float), NULL, &err);
    CL_CHECK_ERROR(err);

    /*
     * Do FFT
     */
    CL_CHECK_ERROR_EX(clFFT_ExecuteInterleaved(queue, fft_plan, 2*N, clFFT_Forward, d_input, d_output, 0, NULL, NULL));
    CL_CHECK_ERROR(clEnqueueReadBuffer(queue, d_output, CL_TRUE, 0, 2*N*sizeof(float), h_output, 0, NULL, NULL));
    print_array(h_output, 16);
    
    CL_CHECK_ERROR_EX(clFFT_ExecuteInterleaved(queue, fft_plan, 2*N, clFFT_Inverse, d_output, d_input, 0, NULL, NULL));
    CL_CHECK_ERROR(clEnqueueReadBuffer(queue, d_input, CL_TRUE, 0, 2*N*sizeof(float), h_output, 0, NULL, NULL));
    for (int i = 0; i < 16; i++)
        h_output[i] /= N;
    print_array(h_output, 16);

    /*
     * Clean up
     */
    CL_CHECK_ERROR(clReleaseMemObject(d_input));
    CL_CHECK_ERROR(clReleaseMemObject(d_output));
    CL_CHECK_ERROR(clReleaseCommandQueue(queue));
    CL_CHECK_ERROR(clReleaseContext(context));
    return 0;
}
